/******************************************************************************
*
*  getaudiodir.cmd
*
*  Create index of MP3-CDs or other similar audio file collections
*
*  Update:  03.04.2007
*  (C)      Marcel Mller 2000-2007
*
******************************************************************************/

IF RxFuncAdd("SysLoadFuncs", "RexxUtil", "SysLoadFuncs") = 0 THEN
   CALL SysLoadFuncs
IF RxFuncAdd("MMLoadFuncs", "RxMMutl", "MMLoadFuncs") = 0 THEN
   CALL MMLoadFuncs

NUMERIC DIGITS 14

monsum = "0 31 59 90 120 151 181 212 243 273 304 334 365"

PARSE ARG vol path

IF (vol = '') THEN
   CALL Error 255, "usage: GETAUDIODIR <drive:> [<destinationpath>]"

freeb = 0
sizeb = 0
vol = STRIP(vol, 'T', '\')
IF RIGHT(vol,1) = ':' THEN DO
   /* whole volume */
   PARSE VALUE SysDriveInfo(FILESPEC('D', vol)) WITH dummy freeb sizeb label
   label = STRIP(label)
   END
 ELSE IF FILESPEC('N', vol) \= '' THEN
   label = FILESPEC('N', vol)
 ELSE IF label = '' THEN
   CALL Error 30, vol" is not a valid source."

IF RIGHT(path,1,'\') \= '\' THEN path = path'\'

olddir = DIRECTORY()
dir = DIRECTORY(vol)
IF dir = '' THEN
   CALL Error 20, "Failed to change to directory "vol
IF RIGHT(dir, 1) \= '\' THEN
   dir = dir'\'
/* get file list */
IF SysFileTree('*', files, 'BOS', '**-*-') \= 0 THEN
   CALL Error 29, "Failed to get file list of"vol"*"
CALL DIRECTORY olddir
offset = LENGTH(dir)+1

CALL fqsort 1, files.0

dst = path||label

sizes = 0

/* read old file content to conserve time */
IF STREAM(dst, 'c', 'open read') = "READY:" THEN DO
   DO WHILE STREAM(dst, 'S') = "READY"
      l = LINEIN(dst)
      IF l = '' THEN ITERATE
      sname = SUBSTR(l, 162)
      p = LASTPOS(' {', sname)
      IF p \= 0 THEN
         sname = LEFT(sname, p-1)
      /*SAY "O: old."sname" . "*/
      old.sname = l
      tmp = SUBSTR(l, 147, 5)
      IF POS('.', tmp) \= 0 THEN
         tmp = tmp * 1024
      sizes = sizes + tmp
      END
   CALL STREAM dst, 'c', 'close'
   CALL SysFileDelete(dst)
   END

/* open destination for writing */
IF STREAM(dst, 'c', 'open write') \= "READY:" THEN
   CALL Error 21, "Failed to open destination file "dst

tmpfile = SysTempFileName(VALUE('TEMP', ,'OS2ENVIRONMENT')'\gadi????.tmp')
dirkeys = ''
/* check files */
DO i = 1 TO files.0
   file = files.i
   sname = SUBSTR(files.i, offset)
   /*SAY "now processing "sname" ..."*/
   /* get file size */
   num = STREAM(file, 'c', 'QUERY SIZE')
   IF num < 128 THEN DO
      IF num = 0 THEN
         /* folder */
         dirkeys = GetKeys(file)
       ELSE
         SAY "TAG of file "file" cannot be read."
      ITERATE
      END
   size = FORMAT(num/1024, 5,0)
   sizes = sizes + size
   /* get time stamp */
   /*SAY "FD: "STREAM(STRIP(file), 'c', 'query datetime')*/
   PARSE VALUE STREAM(STRIP(file), 'c', 'query datetime') WITH month'-'day'-'year hour':'min':'sec
   ts = D2X(ToUnixTime(RIGHT(year+30,2)+1970, month, day, hour, min, sec), 8)
   /* check, if file is known */
   /*SAY "N: old."sname" . "old.sname*/
   IF old.sname \= 'OLD.'sname THEN DO
      l = old.sname
      /*SAY "TC: "SUBSTR(l, 147, 14)" <-> "size ts*/
      IF SUBSTR(l, 147, 14) = size ts THEN DO
         /* replace keys anyway */
         p = LASTPOS(' {', l)
         IF p \= 0 THEN
            l = SUBSTR(l, 1, p-1)
         keys = GetKeys(file, dirkeys)
         IF keys \= '' THEN
            keys = ' {'STRIP(dirkeys keys)'}'
         CALL LINEOUT dst, l||keys
         ITERATE
         END
      END
   /* defaults */
   len = '     '
   rmode = ' '
   rate = '   '
   mode = '  '
   track = '  '
   year = ''
   comment = ''
   title = FILESPEC('N', file)
   IF TRANSLATE(RIGHT(title, 4)) = '.MP3' | TRANSLATE(RIGHT(title, 4)) = '.OGG' THEN
      title = LEFT(title, LENGTH(title)-4)
   PARSE VALUE STRIP(FILESPEC('P', files.i), 'T', '\') WITH artist'\'album
   /* read file type */
   CALL STREAM file, 'c', 'OPEN READ'
   head = CHARIN(file, , 4)
   SELECT
    WHEN head = 'OggS' THEN DO /* ogg vorbis */
      CALL STREAM file, 'c', 'close'
      rmode = '' /* ogg vorbis is allways VBR */
      /* launch vorbiscomment */
      CALL CHAROUT STDOUT, "Scanning "file" ... "
      '@ogginfo "'PCTesc(file)'" >'tmpfile
      IF RC \= 0 THEN DO
         SAY 'Execution of ogginfo returned a bad status ('RC')'
         IF RC = 99 THEN EXIT 99
         END
      /* read result */
       ELSE IF STREAM(tmpfile, 'c', 'open read') \= 'READY:' THEN
         SAY 'Failed to read result from ogginfo ('tmpfile')'
       ELSE DO
         DO WHILE STREAM(tmpfile, 's') = 'READY'
            tmp = LINEIN(tmpfile)
            IF tmp = '' THEN ITERATE
            PARSE VAR tmp key'='val
            key = STRIP2(key)
            SELECT /* comments */
             WHEN key = 'title' | key = 'artist' | key = 'album' | key = 'comment' THEN
               CALL VALUE key, val
             WHEN key = 'tracknumber' THEN
               track = RIGHT(val, 2, '0')
             WHEN key = 'date' THEN
               year = val
             OTHERWISE NOP
               END
            PARSE VAR tmp key': 'val
            key = STRIP2(key)
            SELECT /* file infos */
             WHEN key = 'Playback length' THEN DO
               PARSE VAR val m'm:'s's'
               len = RIGHT(m, 2)':'RIGHT(FORMAT(s, , 0), 2, '0')
               END
             WHEN key = 'Average bitrate' THEN DO
               PARSE VAR val rate dummy
               rate = FORMAT(TRANSLATE(rate, '.', ','), 3, 0)
               END
             OTHERWISE NOP
               END
            END
         CALL STREAM tmpfile, 'c', 'close'
         END
      /* done */
      END

    OTHERWISE /* default: MP3 */
      CALL CHAROUT STDOUT, "Scanning "file" ... "
      /* read tag */
      tag = CHARIN(file, num-127, 128)
      CALL STREAM file, 'c', 'close'
      IF SUBSTR(tag, 1, 3) \= 'TAG' THEN DO
         SAY "file "file" is not an MPEG file or has no TAG information."
         title = FILESPEC('N', file)
         IF TRANSLATE(RIGHT(title, 4)) = '.MP3' THEN
            title = LEFT(title, LENGTH(title)-4)
         PARSE VALUE STRIP(FILESPEC('P', files.i), 'T', '\') WITH artist'\'album
         END
       ELSE DO
         /* split track */
         title = StripStr(SUBSTR(tag, 4, 30))
         artist = StripStr(SUBSTR(tag, 34, 30))
         album = StripStr(SUBSTR(tag, 64, 30))
         year = StripStr(SUBSTR(tag, 94, 4))
         comment = StripStr(SUBSTR(tag, 98, 30))
         IF SUBSTR(tag, 126, 1) = '00'x & SUBSTR(tag, 127, 1) \= '00'x THEN
            track = RIGHT(C2D(SUBSTR(tag, 127, 1)), 2 , '0')
         END
      /* analyze MP3 file */
      err = 0
      '@call Frameanalyzer "'PCTesc(file)'" >'tmpfile' 2>&1'
      IF RC \= 0 THEN DO
         SAY "Failed to analyze "file" - RC: "RC
         IF RC = 99 THEN ITERATE /*EXIT 99*/
         END
       ELSE DO readanalysis = 1 TO 1
         l = LINEIN(tmpfile)
         DO WHILE TRANSLATE(LEFT(l, 5)) \= 'FRAME'
            l = LINEIN(tmpfile)
            IF STREAM(tmpfile, 'S') \= 'READY' THEN DO
               SAY "Failed to read frame analysis result of "file"."
               CALL STREAM tmpfile, 'c', 'close'
               LEAVE readanalysis
               END
            IF ABBREV(TRANSLATE(l), "ERROR") THEN
               err = err +1
            END
         l = LINEIN(tmpfile)
         IF l = '' THEN l = LINEIN(tmpfile)
         PARSE VAR l frames d d layer crc rate samp d d mode ismode d
         IF mode = 'JS' & ismode = 'on' THEN
            mode = 'IS'
         SELECT
          WHEN layer = "I" | layer = 1 THEN
            rmode = "'"
          WHEN layer = "II" | layer = 2 THEN
            rmode = ""
          WHEN layer = "III" | layer = 3 THEN
            rmode = ""
          OTHERWISE NOP
            END
         DO WHILE STREAM(tmpfile, 'S') = "READY"
            l = LINEIN(tmpfile)
            SELECT
             WHEN POS('AVERAGE', TRANSLATE(l)) \= 0 THEN DO
               PARSE VAR l ': ' tmp d
               IF tmp = '' THEN ITERATE
               rate = STRIP(tmp)
               rmode = 'v'
               END
             WHEN POS('TOTAL', TRANSLATE(l)) \= 0 THEN DO
               PARSE VAR l ': ' h':'m':'s d
               len = RIGHT(60*STRIP(h) + m, 2)':'RIGHT(FORMAT(s, , 0), 2, '0')
               END
             OTHERWISE NOP
               END
            END
         CALL STREAM tmpfile, 'c', 'close'
         IF rate > 320 THEN ITERATE
         IF rate \= '' THEN
            rate = FORMAT(rate, 3, 0)
          ELSE
            rate = '   '
         END
      CALL SysFileDelete(tmpfile)
      IF err \= 0 THEN DO
         comment = '$$$'err' 'comment
         CALL LINEOUT STDERR, '$$$ 'err' Errors'
         END
      END
   /* fetch .KEYPHRASES */
   keys = GetKeys(file, dirkeys)
   IF keys \= '' THEN
      keys = ' {'STRIP(dirkeys keys)'}'
   /* write 2 log */
   CALL CHAROUT STDERR, '0d'x
   CALL LINEOUT dst, LEFT(artist,30) LEFT(title,30) LEFT(year,4) len rate||rmode LEFT(mode,2) LEFT(album, 30) LEFT(comment,30) track size ts sname||keys
   END

CALL STREAM dst, 'c', 'close'
CALL SysFileDelete(tmpfile)

/* update overview file */
dst = path"Index"
nindex = 0
/* read */
IF STREAM(dst, 'c', 'open read') = "READY:" THEN DO
   DO FOREVER
      index.nindex = LINEIN(dst)
      IF index.nindex = '' THEN LEAVE
      nindex = nindex +1
      END
   CALL STREAM dst, 'c', 'close'
   END
 ELSE
   nindex = 1
index.0 = "label                           files   total KB    size KB    free KB"
/* modify */
labelt = TRANSLATE(LEFT(label,11))
DO i = 1 TO nindex-1 WHILE TRANSLATE(LEFT(index.i, 11)) < labelt
   END
IF TRANSLATE(LEFT(index.i, 12)) \= labelt THEN DO
   DO j = nindex TO i+1 BY -1
      k = j-1
      index.j = index.k
      END
   nindex = nindex +1
   END
index.i = LEFT(label,30) FORMAT(files.0, 6,0) FORMAT(sizes, 10,0) FORMAT(sizeb % 1024, 10,0) FORMAT(freeb % 1024, 10,0)
/* write */
'@DEL "'PCTesc(dst)'" 2>NUL'
DO i = 0 TO nindex-1
   CALL LINEOUT dst, index.i
   END
CALL STREAM dst, 'c', 'close'


EXIT

/* sort directory tree
   fqsort(index, first, last)
   index tree table index
     1 = tree.1..., 2 = tree.2...
   first, last
     sort range tree.index.first to tree.index.last
   This function calls itself recursively.
*/
fqsort: PROCEDURE EXPOSE files.
   lo = ARG(1)
   hi = ARG(2)
   mi = (hi + lo) %2
   m = TRANSLATE(files.mi, SysGetCollate(), XRANGE('00'x, 'FF'x))
   /* SAY "fqs: "lo hi mi m */
   DO WHILE hi >= lo
      DO WHILE TRANSLATE(files.lo, SysGetCollate(), XRANGE('00'x, 'FF'x)) < m
         lo = lo +1
         END
      DO WHILE TRANSLATE(files.hi, SysGetCollate(), XRANGE('00'x, 'FF'x)) > m
         hi = hi -1
         END
      IF hi >= lo THEN DO
         tmp = files.lo
         files.lo = files.hi
         files.hi = tmp
         lo = lo +1
         hi = hi -1
         END
      END
   IF hi > ARG(1) THEN
      CALL fqsort ARG(1), hi
   IF ARG(2) > lo THEN
      CALL fqsort lo, ARG(2)
   RETURN

/* split multiline EA into peaces
*/
ParseAsEATMVMT: PROCEDURE
   magic = C2D(REVERSE(SUBSTR(ARG(1),1,2)),2)
   IF magic \= -33 THEN
      RETURN '' /* invalid format */
   num = C2D(REVERSE(SUBSTR(ARG(1),5,2)),2)
   r = ''
   EA = SUBSTR(ARG(1), 7)
   DO WHILE num \= 0
      magic = C2D(REVERSE(SUBSTR(EA,1,2)),2)
      IF magic \= -3 THEN
         RETURN ''
      len = C2D(REVERSE(SUBSTR(EA,3,2)),2)
      r = r' 'SUBSTR(EA,5,len)
      EA = SUBSTR(EA,5+len)
      num = num -1
      END
   IF LENGTH(EA) \= 0 THEN
      RETURN ''
   RETURN SUBSTR(r, 2)

/* get keys of file ARG(1) prepending keys ARG(2) if any. */
GetKeys: PROCEDURE
   keys = ''
   IF SysGetEA(ARG(1), ".KEYPHRASES", "KEYS") = 0 & keys \= '' THEN
      keys = ParseAsEATMVMT(keys)
   RETURN STRIP(ARG(2)' 'keys)

/* fetch volume label of drive ARG(1) */
GetVolLabel: PROCEDURE
   di = SysDriveInfo(ARG(1))
   IF di = '' THEN
      CALL Error 10, "cannot access drive "vol
   PARSE VAR di di di di label
   RETURN STRIP(label)

ToUnixtime: PROCEDURE EXPOSE monsum
   year70 = ARG(1) - 1970;
   /* check
   if ( tm->Sec > 59U || tm->Min > 59U || tm->Hour > 23U
       || (unsigned)year70 >= 136U || (tm->Month-1U) > 11U
       || (tm->Day-1) >= monlen[tm->Month-1]
       || (tm->Day == 29 && tm->Month == 2 && (year70&3)) )
      return false;*/
   RETURN FORMAT(ARG(6) + 60*ARG(5) + 3600*ARG(4) + 86400*(ARG(3)-1 + WORD(monsum, ARG(2)) + (ARG(2) >= 3 & year70//4 == 2) + 365*year70 + (year70+1)%4),,0,,20)

StripStr: PROCEDURE
   p = POS('0'x, ARG(1)'0'x)
   RETURN Cp2DOS(STRIP(LEFT(ARG(1), p-1)))

PCTesc: PROCEDURE
   tmp = ARG(1)
   p = POS('%', tmp)
   DO WHILE p \= 0
      tmp = INSERT('%', tmp, p)
      p = POS('%', tmp, p+2)
      END
   RETURN tmp

Cp2DOS: /* CP translation ISO to current CP */
   IF RxFuncQuery("MMTranslateCP") \= 0 THEN
      RETURN ARG(1)  /* can't help it */
   RETURN MMTranslateCp(ARG(1), 1004, , '')

Strip2: PROCEDURE
   tmp = ARG(1)
   opt = TRANSLATE(LEFT(ARG(2), 1, 'B'))
   IF ARG(3, 'e') THEN
      chr = ARG(2)
    ELSE
      chr = '2009'x
   /* strip front */
   IF TRANSLATE(opt, 'L', 'B') = 'L' THEN DO
      p = VERIFY(tmp, chr)
      IF p = 0 THEN
         RETURN ''
       ELSE
         tmp = SUBSTR(tmp, p)
      END
   /* strip back */
   IF TRANSLATE(opt, 'T', 'B') = 'T' THEN DO
      p = VERIFY(REVERSE(tmp), chr)
      RETURN LEFT(tmp, LENGTH(tmp) - p +1)
      END
   RETURN tmp

Error: PROCEDURE /* raise error */
   SAY ARG(2)
   EXIT ARG(1)

